!***********************************************************************
!*                   GNU Lesser General Public License
!*
!* This file is part of the FV3 dynamical core.
!*
!* The FV3 dynamical core is free software: you can redistribute it
!* and/or modify it under the terms of the
!* GNU Lesser General Public License as published by the
!* Free Software Foundation, either version 3 of the License, or
!* (at your option) any later version.
!*
!* The FV3 dynamical core is distributed in the hope that it will be
!* useful, but WITHOUT ANYWARRANTY; without even the implied warranty
!* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
!* See the GNU General Public License for more details.
!*
!* You should have received a copy of the GNU Lesser General Public
!* License along with the FV3 dynamical core.
!* If not, see <http://www.gnu.org/licenses/>.
!***********************************************************************
module atmosphere_mod
#include <fms_platform.h>

!-----------------------------------------------------------------------
!
! Interface for Cubed_Sphere fv dynamical core
!
!-----------------------------------------------------------------------

!-----------------
! FMS modules:
!-----------------
use atmos_co2_mod,      only: atmos_co2_rad, co2_radiation_override
use block_control_mod,  only: block_control_type
use constants_mod,      only: cp_air, rdgas, grav, rvgas, kappa, pstd_mks
use time_manager_mod,   only: time_type, get_time, set_time, operator(+)
use fms_mod,            only: file_exist, open_namelist_file,    &
                              close_file, error_mesg, FATAL,     &
                              check_nml_error, stdlog,           &
                              write_version_number,              &
                              mpp_pe, mpp_root_pe, set_domain,   &
                              mpp_clock_id, mpp_clock_begin,     &
                              mpp_clock_end, CLOCK_SUBCOMPONENT, &
                              clock_flag_default, nullify_domain
use mpp_mod,            only: mpp_error, FATAL, NOTE, input_nml_file, &
                              mpp_npes, mpp_get_current_pelist, &
                              mpp_set_current_pelist, stdout, &
                              mpp_pe, mpp_chksum
use mpp_domains_mod,    only: domain2d
use xgrid_mod,          only: grid_box_type
!miz
use diag_manager_mod,   only: register_diag_field, send_data
use field_manager_mod,  only: MODEL_ATMOS
use tracer_manager_mod, only: get_tracer_index,&
                              get_number_tracers, &
                              get_tracer_names, NO_TRACER
use physics_driver_mod, only: surf_diff_type
use physics_types_mod,  only: physics_type, &
                              physics_tendency_type
use radiation_types_mod,only: radiation_type, compute_g_avg
use atmos_cmip_diag_mod,only: atmos_cmip_diag_init, &
                              register_cmip_diag_field_3d, &
                              send_cmip_data_3d, cmip_diag_id_type, &
                              query_cmip_diag_id
#ifndef use_AM3_physics
use  atmos_global_diag_mod, only: atmos_global_diag_init, &
                                  atmos_global_diag_end
#endif

!-----------------
! FV core modules:
!-----------------
use fv_arrays_mod,      only: fv_atmos_type
use fv_control_mod,     only: fv_control_init, fv_end, ngrids
use fv_eta_mod,         only: get_eta_level
use fv_io_mod,          only: fv_io_register_nudge_restart
use fv_dynamics_mod,    only: fv_dynamics
use fv_nesting_mod,     only: twoway_nesting
use fv_diagnostics_mod, only: fv_diag_init, fv_diag, fv_time, prt_maxmin
use fv_cmip_diag_mod,   only: fv_cmip_diag_init, fv_cmip_diag, fv_cmip_diag_end
use fv_restart_mod,     only: fv_restart, fv_write_restart
use fv_timing_mod,      only: timing_on, timing_off
use fv_mp_mod,          only: switch_current_Atm
use fv_sg_mod,          only: fv_subgrid_z
use fv_update_phys_mod, only: fv_update_phys
#if defined (ATMOS_NUDGE)
use atmos_nudge_mod,    only: atmos_nudge_init, atmos_nudge_end
#elif defined (CLIMATE_NUDGE)
use fv_climate_nudge_mod,only: fv_climate_nudge_init,fv_climate_nudge_end
#elif defined (ADA_NUDGE)
use fv_ada_nudge_mod,   only: fv_ada_nudge_init, fv_ada_nudge_end
#else
use fv_nwp_nudge_mod,   only: fv_nwp_nudge_init, fv_nwp_nudge_end, do_adiabatic_init
use amip_interp_mod,    only: forecast_mode
#endif

use mpp_domains_mod, only:  mpp_get_data_domain, mpp_get_compute_domain

implicit none
private

!--- driver routines
public :: atmosphere_init, atmosphere_end, atmosphere_restart, &
          atmosphere_dynamics, atmosphere_state_update

!--- utility routines
public :: atmosphere_resolution, atmosphere_boundary, &
          atmosphere_grid_center, atmosphere_domain, &
          atmosphere_cell_area, atmosphere_control_data, &
          atmosphere_pref, &
          get_atmosphere_axes, get_bottom_mass, &
          get_bottom_wind, get_stock_pe, &
          set_atmosphere_pelist, reset_atmos_tracers

!--- physics/radiation data exchange routines
public :: atmos_radiation_driver_inputs, atmos_physics_driver_inputs

!-----------------------------------------------------------------------

! version number of this module
! Include variable "version" to be written to log file.
#include<file_version.h>
character(len=7)   :: mod_name = 'atmos'

!---- private data ----
  type (time_type) :: Time_step_atmos
  public Atm

  !These are convenience variables for local use only, and are set to values in Atm%
  real    :: dt_atmos
  real    :: zvir
  integer :: npx, npy, npz, ncnst, pnats
  integer :: isc, iec, jsc, jec
  integer :: isd, ied, jsd, jed
  integer :: nq                       ! transported tracers
  integer :: sec, seconds, days
  integer :: id_dynam, id_fv_diag, id_subgridz
  logical :: cold_start = .false.       ! read in initial condition

  integer, dimension(:), allocatable :: id_tracerdt_dyn
  integer :: num_tracers = 0
!miz
  !Diagnostics
  integer :: id_tdt_dyn, id_qdt_dyn, id_qldt_dyn, id_qidt_dyn, id_qadt_dyn
  logical :: used
  character(len=64) :: field
  real, allocatable :: ttend(:,:,:)
  real, allocatable :: qtendyyf(:,:,:,:)
  real, allocatable :: qtend(:,:,:,:)
  real              :: mv = -1.e10
!miz
  type(cmip_diag_id_type) :: ID_tnta, ID_tnhusa

  integer :: mytile = 1
  integer :: p_split = 1
  integer, allocatable :: pelist(:)
  logical, allocatable :: grids_on_this_pe(:)
  integer :: this_grid
  type(fv_atmos_type), allocatable, target :: Atm(:)

  integer :: id_udt_dyn, id_vdt_dyn

  real, parameter:: w0_big = 60.  ! to prevent negative w-tracer diffusion

!---dynamics tendencies for use in fv_subgrid_z and during fv_update_phys
  real, allocatable, dimension(:,:,:)   :: u_dt, v_dt, t_dt
  real, allocatable, dimension(:,:,:,:) :: q_dt
  real, allocatable :: pref(:,:), dum1d(:)

!---need to define do_adiabatic_init to satisfy a reference when nwp_nudge is not the default
#if defined(ATMOS_NUDGE) || defined(CLIMATE_NUDGE) || defined(ADA_NUDGE)
   logical :: do_adiabatic_init
#endif

  integer, parameter :: kind_phys=8

contains



 subroutine atmosphere_init (Time_init, Time, Time_step, Surf_diff, Grid_box)
   type (time_type),      intent(in)    :: Time_init, Time, Time_step
   type(surf_diff_type),  intent(inout) :: Surf_diff
   type(grid_box_type),   intent(inout) :: Grid_box

!--- local variables ---
   integer :: i, n
   integer :: itrac
   logical :: do_atmos_nudge
   character(len=32) :: tracer_name, tracer_units
   real :: ps1, ps2

                    call timing_on('ATMOS_INIT')
   allocate(pelist(mpp_npes()))
   call mpp_get_current_pelist(pelist)

   call get_number_tracers(MODEL_ATMOS, num_prog= num_tracers)

   zvir = rvgas/rdgas - 1.

!---- compute physics/atmos time step in seconds ----

   Time_step_atmos = Time_step
   call get_time (Time_step_atmos, sec)
   dt_atmos = real(sec)

!----- initialize FV dynamical core -----
   !NOTE do we still need the second file_exist call?
   cold_start = (.not.file_exist('INPUT/fv_core.res.nc') .and. .not.file_exist('INPUT/fv_core.res.tile1.nc'))

   call fv_control_init( Atm, dt_atmos, this_grid, grids_on_this_pe, p_split )  ! allocates Atm components

   do n=1,ngrids
      if (grids_on_this_pe(n)) mytile = n
   enddo

!----- write version and namelist to log file -----
   call write_version_number ( 'COUPLED/ATMOSPHERE_MOD', version )

!-----------------------------------

   npx   = Atm(mytile)%npx
   npy   = Atm(mytile)%npy
   npz   = Atm(mytile)%npz
   ncnst = Atm(mytile)%ncnst
   pnats = Atm(mytile)%flagstruct%pnats

   isc = Atm(mytile)%bd%isc
   iec = Atm(mytile)%bd%iec
   jsc = Atm(mytile)%bd%jsc
   jec = Atm(mytile)%bd%jec

   isd = isc - Atm(mytile)%bd%ng
   ied = iec + Atm(mytile)%bd%ng
   jsd = jsc - Atm(mytile)%bd%ng
   jed = jec + Atm(mytile)%bd%ng

   nq = ncnst-pnats

   ! Allocate grid variables to be used to calculate gradient in 2nd order flux exchange
   ! This data is only needed for the COARSEST grid.
   call switch_current_Atm(Atm(mytile))

   allocate(Grid_box%dx    (   isc:iec  , jsc:jec+1))
   allocate(Grid_box%dy    (   isc:iec+1, jsc:jec  ))
   allocate(Grid_box%area  (   isc:iec  , jsc:jec  ))
   allocate(Grid_box%edge_w(              jsc:jec+1))
   allocate(Grid_box%edge_e(              jsc:jec+1))
   allocate(Grid_box%edge_s(   isc:iec+1           ))
   allocate(Grid_box%edge_n(   isc:iec+1           ))
   allocate(Grid_box%en1   (3, isc:iec  , jsc:jec+1))
   allocate(Grid_box%en2   (3, isc:iec+1, jsc:jec  ))
   allocate(Grid_box%vlon  (3, isc:iec  , jsc:jec  ))
   allocate(Grid_box%vlat  (3, isc:iec  , jsc:jec  ))
   Grid_box%dx    (   isc:iec  , jsc:jec+1) = Atm(mytile)%gridstruct%dx    (   isc:iec,   jsc:jec+1)
   Grid_box%dy    (   isc:iec+1, jsc:jec  ) = Atm(mytile)%gridstruct%dy    (   isc:iec+1, jsc:jec  )
   Grid_box%area  (   isc:iec  , jsc:jec  ) = Atm(mytile)%gridstruct%area  (   isc:iec  , jsc:jec  )
   Grid_box%edge_w(              jsc:jec+1) = Atm(mytile)%gridstruct%edge_w(              jsc:jec+1)
   Grid_box%edge_e(              jsc:jec+1) = Atm(mytile)%gridstruct%edge_e(              jsc:jec+1)
   Grid_box%edge_s(   isc:iec+1           ) = Atm(mytile)%gridstruct%edge_s(   isc:iec+1)
   Grid_box%edge_n(   isc:iec+1           ) = Atm(mytile)%gridstruct%edge_n(   isc:iec+1)
   Grid_box%en1   (:, isc:iec  , jsc:jec+1) = Atm(mytile)%gridstruct%en1   (:, isc:iec  , jsc:jec+1)
   Grid_box%en2   (:, isc:iec+1, jsc:jec  ) = Atm(mytile)%gridstruct%en2   (:, isc:iec+1, jsc:jec  )
   do i = 1,3
     Grid_box%vlon  (i, isc:iec  , jsc:jec  ) = Atm(mytile)%gridstruct%vlon  (isc:iec ,  jsc:jec, i )
     Grid_box%vlat  (i, isc:iec  , jsc:jec  ) = Atm(mytile)%gridstruct%vlat  (isc:iec ,  jsc:jec, i )
   enddo

!----- allocate and zero out the dynamics (and accumulated) tendencies
   allocate( u_dt(isd:ied,jsd:jed,npz), &
             v_dt(isd:ied,jsd:jed,npz), &
             t_dt(isc:iec,jsc:jec,npz), &
             q_dt(isc:iec,jsc:jec,npz,nq) )
!--- allocate pref
    allocate(pref(npz+1,2), dum1d(npz+1))

   call set_domain ( Atm(mytile)%domain )
   call fv_restart(Atm(mytile)%domain, Atm, dt_atmos, seconds, days, cold_start, Atm(mytile)%gridstruct%grid_type, mytile)

   fv_time = Time

!----- initialize atmos_axes and fv_dynamics diagnostics
       !I've had trouble getting this to work with multiple grids at a time; worth revisiting?
   call fv_diag_init(Atm(mytile:mytile), Atm(mytile)%atmos_axes, Time, npx, npy, npz, Atm(mytile)%flagstruct%p_ref)

!---------- reference profile -----------
    ps1 = 101325.
    ps2 =  81060.
    pref(npz+1,1) = ps1
    pref(npz+1,2) = ps2
    call get_eta_level ( npz, ps1, pref(1,1), dum1d, Atm(mytile)%ak, Atm(mytile)%bk )
    call get_eta_level ( npz, ps2, pref(1,2), dum1d, Atm(mytile)%ak, Atm(mytile)%bk )

!---- initialize cmip diagnostic output ----
   call atmos_cmip_diag_init   ( Atm(mytile)%ak, Atm(mytile)%bk, pref(1,1), Atm(mytile)%atmos_axes, Time )
#ifndef use_AM3_physics
   call atmos_global_diag_init ( Atm(mytile)%atmos_axes, Atm(mytile)%gridstruct%area(isc:iec,jsc:jec) )
#endif
   call fv_cmip_diag_init      ( Atm(mytile:mytile), Atm(mytile)%atmos_axes, Time )

!--- initialize nudging module ---
#if defined (ATMOS_NUDGE)
    call atmos_nudge_init ( Time, Atm(mytile)%atmos_axes(1:3), flag=do_atmos_nudge )
    if ( do_atmos_nudge .and. Atm(mytile)%flagstruct%nudge ) then
         call mpp_error(NOTE, 'Code compiled with atmospheric nudging, but fv_core_nml nudge is also set to .true.')
    elseif ( do_atmos_nudge) then
         call mpp_error(NOTE, 'Code compiled with and using atmospheric nudging')
    endif
    Atm(mytile)%flagstruct%nudge = do_atmos_nudge
#elif defined (CLIMATE_NUDGE)
    call fv_climate_nudge_init ( Time, Atm(mytile)%atmos_axes(1:3), flag=do_atmos_nudge )
    if ( do_atmos_nudge .and. Atm(1)%flagstruct%nudge ) then
         call mpp_error(NOTE, 'Code compiled with climate nudging, but fv_core_nml nudge is also set to .true.')
    elseif ( do_atmos_nudge ) then
         call mpp_error(NOTE, 'Code compiled with and using climate nudging')
    endif
    Atm(mytile)%flagstruct%nudge = do_atmos_nudge
#elif defined (ADA_NUDGE)
    if ( Atm(1)%flagstruct%nudge ) then
        call fv_ada_nudge_init( Time, Atm(mytile)%atmos_axes, npz, zvir, Atm(1)%ak, Atm(1)%bk, Atm(1)%ts, &
           Atm(1)%phis, Atm(1)%gridstruct, Atm(1)%ks, Atm(1)%npx, Atm(1)%neststruct, Atm(1)%bd, Atm(1)%domain)
        call mpp_error(NOTE, 'ADA nudging is active')
     endif
#else
   !Only do nudging on coarse grid for now
   if ( Atm(mytile)%flagstruct%nudge ) then
      call fv_nwp_nudge_init( Time, Atm(mytile)%atmos_axes, npz, zvir, Atm(1)%ak, Atm(1)%bk, Atm(1)%ts, &
           Atm(1)%phis, Atm(1)%gridstruct, Atm(1)%ks, Atm(1)%npx, Atm(1)%neststruct, Atm(1)%bd)
        call mpp_error(NOTE, 'NWP nudging is active')
   endif
#endif

! This call needs to be separate from the register nudging restarts after initialization
   call fv_io_register_nudge_restart ( Atm )

!uc/vc only need be same on coarse grid? However BCs do need to be the same

   !This appears to all be diagnostics through the end of this routine,
   !and so for now we will only define for the coarsest grid

!miz
   !---allocate id_tracer_*
   allocate (id_tracerdt_dyn    (num_tracers))
   if ( Atm(mytile)%flagstruct%write_3d_diags) then
      id_udt_dyn    =register_diag_field(mod_name,'udt_dyn',    Atm(mytile)%atmos_axes(1:3),  &
           Time,'udt_dyn',    'm/s/s', missing_value=mv)
      id_vdt_dyn    =register_diag_field(mod_name,'vdt_dyn',    Atm(mytile)%atmos_axes(1:3),  &
           Time,'vdt_dyn',    'm/s/s', missing_value=mv)
      id_tdt_dyn    =register_diag_field(mod_name,'tdt_dyn',    Atm(mytile)%atmos_axes(1:3),  &
           Time,'tdt_dyn',    'K/s', missing_value=mv)
      id_qdt_dyn    =register_diag_field(mod_name,'qdt_dyn',    Atm(mytile)%atmos_axes(1:3),  &
           Time,'qdt_dyn',    'kg/kg/s', missing_value=mv)
      id_qldt_dyn   =register_diag_field(mod_name,'qldt_dyn',   Atm(mytile)%atmos_axes(1:3),  &
           Time,'qldt_dyn',   'kg/kg/s', missing_value=mv)
      id_qidt_dyn   =register_diag_field(mod_name,'qidt_dyn',   Atm(mytile)%atmos_axes(1:3),  &
           Time,'qidt_dyn',   'kg/kg/s', missing_value=mv)
      id_qadt_dyn   =register_diag_field(mod_name,'qadt_dyn',   Atm(mytile)%atmos_axes(1:3),  &
           Time,'qadt_dyn',   '1/s', missing_value=mv)
      !--- register cmip tendency fields ---
      ID_tnta = register_cmip_diag_field_3d (mod_name, 'tnta', Time, &
           'Tendency of Air Temperature due to Advection', 'K s-1', &
           standard_name='tendency_of_air_temperature_due_to_advection')
      ID_tnhusa = register_cmip_diag_field_3d (mod_name, 'tnhusa', Time, &
           'Tendency of Specific Humidity due to Advection', 's-1', &
           standard_name='tendency_of_specific_humidity_due_to_advection')

      !---loop for tracers
      do itrac = 1, num_tracers
         call get_tracer_names (MODEL_ATMOS, itrac, name = tracer_name, units = tracer_units)
         if (get_tracer_index(MODEL_ATMOS,tracer_name)>0) then
            id_tracerdt_dyn(itrac) = register_diag_field(mod_name, TRIM(tracer_name)//'dt_dyn',  &
                 Atm(mytile)%atmos_axes(1:3),Time,                       &
                 TRIM(tracer_name)//' total tendency from advection',    &
                 TRIM(tracer_units)//'/s', missing_value = mv)
         endif
      enddo
   endif
   if (any(id_tracerdt_dyn(:)>0)) allocate(qtendyyf(isc:iec, jsc:jec,1:npz,num_tracers))
   if ( id_tdt_dyn>0 .or. query_cmip_diag_id(ID_tnta) ) allocate(ttend(isc:iec, jsc:jec, 1:npz))
   if ( any((/ id_qdt_dyn, id_qldt_dyn, id_qidt_dyn, id_qadt_dyn /) > 0) .or. &
        query_cmip_diag_id(ID_tnhusa) )  allocate(qtend(isc:iec, jsc:jec, 1:npz, 4))
!miz

!  --- initialize clocks for dynamics, physics_down and physics_up
   id_dynam     = mpp_clock_id ('FV dy-core',  flags = clock_flag_default, grain=CLOCK_SUBCOMPONENT )
   id_subgridz  = mpp_clock_id ('FV subgrid_z',flags = clock_flag_default, grain=CLOCK_SUBCOMPONENT )
   id_fv_diag   = mpp_clock_id ('FV Diag',     flags = clock_flag_default, grain=CLOCK_SUBCOMPONENT )

   n = mytile
   call switch_current_Atm(Atm(n))

 end subroutine atmosphere_init



 subroutine atmosphere_dynamics ( Time, surf_diff )
   type(time_type),intent(in) :: Time
   integer :: itrac, n, psc
   integer :: k, w_diff, nt_dyn
   type(surf_diff_type), intent(inout):: surf_diff

!---- Call FV dynamics -----

   call mpp_clock_begin (id_dynam)
!miz
#ifndef use_AM3_physics
   Surf_diff%ddp_dyn(:,:,:) = Atm(mytile)%delp(isc:iec, jsc:jec, :)
   Surf_diff%tdt_dyn(:,:,:) = Atm(mytile)%pt(isc:iec, jsc:jec, :)
   Surf_diff%qdt_dyn(:,:,:) = Atm(mytile)%q (isc:iec, jsc:jec, :, 1) + &
                              Atm(mytile)%q (isc:iec, jsc:jec, :, 2) + &
                              Atm(mytile)%q (isc:iec, jsc:jec, :, 3)
#endif

!miz[M d0
   if ( id_tdt_dyn>0 .or. query_cmip_diag_id(ID_tnta) ) ttend(:, :, :) = Atm(mytile)%pt(isc:iec, jsc:jec, :)
   if ( any((/ id_qdt_dyn, id_qldt_dyn, id_qidt_dyn, id_qadt_dyn /) > 0) .or. &
        query_cmip_diag_id(ID_tnhusa) ) qtend(:, :, :, :) = Atm(mytile)%q (isc:iec, jsc:jec, :, :)
!miz
   do itrac = 1, num_tracers
     if (id_tracerdt_dyn (itrac) >0 ) &
            qtendyyf(:,:,:,itrac) = Atm(mytile)%q(isc:iec,jsc:jec,:,itrac)
   enddo

   n = mytile
   do psc=1,abs(p_split)
                    call timing_on('fv_dynamics')
!uc/vc only need be same on coarse grid? However BCs do need to be the same
     call fv_dynamics(npx, npy, npz, nq, Atm(n)%ng, dt_atmos/real(abs(p_split)),&
                      Atm(n)%flagstruct%consv_te, Atm(n)%flagstruct%fill,  &
                      Atm(n)%flagstruct%reproduce_sum, kappa, cp_air, zvir,&
                      Atm(n)%ptop, Atm(n)%ks, nq,                          &
                      Atm(n)%flagstruct%n_split, Atm(n)%flagstruct%q_split,&
                      Atm(n)%u, Atm(n)%v, Atm(n)%w, Atm(n)%delz,           &
                      Atm(n)%flagstruct%hydrostatic,                       &
                      Atm(n)%pt, Atm(n)%delp, Atm(n)%q, Atm(n)%ps,         &
                      Atm(n)%pe, Atm(n)%pk, Atm(n)%peln,                   &
                      Atm(n)%pkz, Atm(n)%phis, Atm(n)%q_con,               &
                      Atm(n)%omga, Atm(n)%ua, Atm(n)%va, Atm(n)%uc,        &
                      Atm(n)%vc, Atm(n)%ak, Atm(n)%bk, Atm(n)%mfx,         &
                      Atm(n)%mfy, Atm(n)%cx, Atm(n)%cy, Atm(n)%ze0,        &
                      Atm(n)%flagstruct%hybrid_z,                          &
                      Atm(n)%gridstruct, Atm(n)%flagstruct,                &
                      Atm(n)%neststruct, Atm(n)%idiag, Atm(n)%bd,          &
                      Atm(n)%parent_grid, Atm(n)%domain, Atm(n)%inline_mp)

     call timing_off('fv_dynamics')

    if (ngrids > 1 .and. (psc < p_split .or. p_split < 0)) then
       call timing_on('TWOWAY_UPDATE')
       call twoway_nesting(Atm, ngrids, grids_on_this_pe, zvir, fv_time, mytile)
       call timing_off('TWOWAY_UPDATE')
    endif

    end do !p_split
    call mpp_clock_end (id_dynam)

!miz
#ifndef use_AM3_physics
   Surf_diff%ddp_dyn(:,:,:) =(Atm(mytile)%delp(isc:iec,jsc:jec,:)-Surf_diff%ddp_dyn(:,:,:))/dt_atmos
   Surf_diff%tdt_dyn(:,:,:) =(Atm(mytile)%pt(isc:iec,jsc:jec,:)  -Surf_diff%tdt_dyn(:,:,:))/dt_atmos
   Surf_diff%qdt_dyn(:,:,:) =(Atm(mytile)%q (isc:iec,jsc:jec,:,1) + &
                              Atm(mytile)%q (isc:iec,jsc:jec,:,2) + &
                              Atm(mytile)%q (isc:iec,jsc:jec,:,3) - Surf_diff%qdt_dyn(:,:,:))/dt_atmos
#endif
!miz
   if ( id_udt_dyn>0 )  used = send_data( id_udt_dyn, 2.0/dt_atmos*Atm(mytile)%ua(isc:iec,jsc:jec,:), Time)
   if ( id_vdt_dyn>0 )  used = send_data( id_vdt_dyn, 2.0/dt_atmos*Atm(mytile)%va(isc:iec,jsc:jec,:), Time)
   if ( id_tdt_dyn>0 .or. query_cmip_diag_id(ID_tnta) ) then
        ttend = (Atm(mytile)%pt(isc:iec, jsc:jec, :)   - ttend(:, :, :   ))/dt_atmos
        if (id_tdt_dyn>0)                used = send_data(id_tdt_dyn,  ttend(:,:,:),   Time)
        if (query_cmip_diag_id(ID_tnta)) used = send_cmip_data_3d (ID_tnta, ttend(:,:,:), Time)
   endif

   if ( any((/ id_qdt_dyn, id_qldt_dyn, id_qidt_dyn, id_qadt_dyn /) > 0) .or.  query_cmip_diag_id(ID_tnhusa) ) then
        qtend = (Atm(mytile)%q (isc:iec, jsc:jec, :, :)- qtend(:, :, :, :))/dt_atmos
        if (id_qdt_dyn  > 0) used = send_data(id_qdt_dyn,  qtend(:,:,:,1), Time)
        if (id_qldt_dyn > 0) used = send_data(id_qldt_dyn, qtend(:,:,:,2), Time)
        if (id_qidt_dyn > 0) used = send_data(id_qidt_dyn, qtend(:,:,:,3), Time)
        if (id_qadt_dyn > 0) used = send_data(id_qadt_dyn, qtend(:,:,:,4), Time)
        if (query_cmip_diag_id(ID_tnhusa)) used = send_cmip_data_3d (ID_tnhusa, qtend(:,:,:,1), Time)
   endif
!miz

   do itrac = 1, num_tracers
     if(id_tracerdt_dyn(itrac)>0) then
       qtendyyf(:,:,:,itrac) = (Atm(mytile)%q (isc:iec, jsc:jec, :,itrac)-  &
                                        qtendyyf(:,:,:,itrac))/dt_atmos
       used = send_data(id_tracerdt_dyn(itrac), qtendyyf(:,:,:,itrac), &
                                                           Time)
     endif
   enddo

!-----------------------------------------------------
!--- COMPUTE SUBGRID Z
!-----------------------------------------------------
!--- zero out tendencies
    call mpp_clock_begin (id_subgridz)
    u_dt(:,:,:)   = 0.
    v_dt(:,:,:)   = 0.
    t_dt(:,:,:)   = 0.
    q_dt(:,:,:,:) = 0.

    w_diff = get_tracer_index (MODEL_ATMOS, 'w_diff' )
    if ( Atm(n)%flagstruct%fv_sg_adj > 0 ) then
      nt_dyn = nq
      if ( w_diff /= NO_TRACER ) then
        nt_dyn = nq - 1
      endif
      call fv_subgrid_z(isd, ied, jsd, jed, isc, iec, jsc, jec, Atm(n)%npz, &
                        nt_dyn, dt_atmos, Atm(n)%flagstruct%fv_sg_adj,      &
                        Atm(n)%flagstruct%nwat, Atm(n)%delp, Atm(n)%pe,     &
                        Atm(n)%peln, Atm(n)%pkz, Atm(n)%pt, Atm(n)%q,       &
                        Atm(n)%ua, Atm(n)%va, Atm(n)%flagstruct%hydrostatic,&
                        Atm(n)%w, Atm(n)%delz, u_dt, v_dt, t_dt, q_dt,      &
                        Atm(n)%flagstruct%n_sponge)
    endif

    if ( .not. Atm(n)%flagstruct%hydrostatic .and. w_diff /= NO_TRACER ) then
!$OMP parallel do default (none) &
!$OMP              shared (isc, iec, jsc, jec, w_diff, n, Atm, q_dt) &
!$OMP             private (k)
       do k=1, Atm(n)%npz
          Atm(n)%q(isc:iec,jsc:jec,k,w_diff) = Atm(n)%w(isc:iec,jsc:jec,k) + w0_big
          q_dt(:,:,k,w_diff) = 0.
        enddo
    endif

   call mpp_clock_end (id_subgridz)

 end subroutine atmosphere_dynamics


 subroutine atmosphere_end (Time, Grid_box )!rab, Radiation, Physics)
   type (time_type),      intent(in)    :: Time
   type(grid_box_type),   intent(inout) :: Grid_box
!rab   type (radiation_type), intent(inout) :: Radiation
!rab   type (physics_type),   intent(inout) :: Physics

  ! initialize domains for writing global physics data
   call set_domain ( Atm(mytile)%domain )


!--- end nudging module ---
#if defined (ATMOS_NUDGE)
   if ( Atm(mytile)%flagstruct%nudge ) call atmos_nudge_end
#elif defined (CLIMATE_NUDGE)
   if ( Atm(mytile)%flagstruct%nudge ) call fv_climate_nudge_end
#elif defined (ADA_NUDGE)
   if ( Atm(mytile)%flagstruct%nudge ) call fv_ada_nudge_end
#else
   if ( Atm(mytile)%flagstruct%nudge ) call fv_nwp_nudge_end
#endif

#ifndef use_AM3_physics
   call atmos_global_diag_end
#endif
   call fv_cmip_diag_end
   call nullify_domain ( )
   call fv_end(Atm, mytile)
   deallocate (Atm)

   deallocate( u_dt, v_dt, t_dt, q_dt, pref, dum1d )

 end subroutine atmosphere_end



  !#######################################################################
  ! <SUBROUTINE NAME="atmosphere_restart">
  ! <DESCRIPTION>
  !  Write out restart files registered through register_restart_file
  ! </DESCRIPTION>
  subroutine atmosphere_restart(timestamp)
    character(len=*),  intent(in) :: timestamp

    call fv_write_restart(Atm(mytile), timestamp)

  end subroutine atmosphere_restart
  ! </SUBROUTINE>


 subroutine atmosphere_resolution (i_size, j_size, global)
   integer, intent(out)          :: i_size, j_size
   logical, intent(in), optional :: global
   logical :: local

   local = .true.
   if( PRESENT(global) ) local = .NOT.global

   if( local ) then
       i_size = iec - isc + 1
       j_size = jec - jsc + 1
   else
       i_size = npx - 1
       j_size = npy - 1
   end if

 end subroutine atmosphere_resolution


 subroutine atmosphere_pref (p_ref)
   real, dimension(:,:), intent(inout) :: p_ref

   p_ref = pref

 end subroutine atmosphere_pref

 subroutine atmosphere_control_data (i1, i2, j1, j2, kt, p_hydro, hydro, do_uni_zfull) !miz
   integer, intent(out)           :: i1, i2, j1, j2, kt
   logical, intent(out), optional :: p_hydro, hydro, do_uni_zfull !miz
   i1 = Atm(mytile)%bd%isc
   i2 = Atm(mytile)%bd%iec
   j1 = Atm(mytile)%bd%jsc
   j2 = Atm(mytile)%bd%jec
   kt = Atm(mytile)%npz

   if (present(p_hydro)) p_hydro = Atm(mytile)%flagstruct%phys_hydrostatic
   if (present(  hydro))   hydro = Atm(mytile)%flagstruct%hydrostatic
   if (present(do_uni_zfull)) do_uni_zfull = Atm(mytile)%flagstruct%do_uni_zfull

 end subroutine atmosphere_control_data


 subroutine atmosphere_cell_area  (area_out)
   real, dimension(:,:),  intent(out)          :: area_out

   area_out(1:iec-isc+1, 1:jec-jsc+1) =  Atm(mytile)%gridstruct%area (isc:iec,jsc:jec)

 end subroutine atmosphere_cell_area



 subroutine atmosphere_grid_center (lon, lat)
!---------------------------------------------------------------
!    returns the longitude and latitude cell centers
!---------------------------------------------------------------
    real(kind=kind_phys), intent(out) :: lon(:,:), lat(:,:)   ! Unit: radian
! Local data:
    integer i,j

    do j=jsc,jec
       do i=isc,iec
          lon(i-isc+1,j-jsc+1) = Atm(mytile)%gridstruct%agrid_64(i,j,1)
          lat(i-isc+1,j-jsc+1) = Atm(mytile)%gridstruct%agrid_64(i,j,2)
       enddo
    end do

 end subroutine atmosphere_grid_center



 subroutine atmosphere_boundary (blon, blat, global)
!---------------------------------------------------------------
!    returns the longitude and latitude grid box edges
!    for either the local PEs grid (default) or the global grid
!---------------------------------------------------------------
    real,    intent(out) :: blon(:,:), blat(:,:)   ! Unit: radian
    logical, intent(in), optional :: global
! Local data:
    integer i,j

    if( PRESENT(global) ) then
      if (global) call mpp_error(FATAL, '==> global grid is no longer available &
                               & in the Cubed Sphere')
    endif

    do j=jsc,jec+1
       do i=isc,iec+1
          blon(i-isc+1,j-jsc+1) = Atm(mytile)%gridstruct%grid(i,j,1)
          blat(i-isc+1,j-jsc+1) = Atm(mytile)%gridstruct%grid(i,j,2)
       enddo
    end do

 end subroutine atmosphere_boundary


 subroutine set_atmosphere_pelist ()
   call mpp_set_current_pelist(Atm(mytile)%pelist, no_sync=.TRUE.)
 end subroutine set_atmosphere_pelist


 subroutine atmosphere_domain ( fv_domain )
   type(domain2d), intent(out) :: fv_domain
!  returns the domain2d variable associated with the coupling grid
!  note: coupling is done using the mass/temperature grid with no halos

   fv_domain = Atm(mytile)%domain_for_coupler

 end subroutine atmosphere_domain



 subroutine get_atmosphere_axes ( axes )
   integer, intent(out) :: axes (:)

!----- returns the axis indices for the atmospheric (mass) grid -----
   if ( size(axes(:)) < 0 .or. size(axes(:)) > 4 ) call error_mesg (    &
                               'get_atmosphere_axes in atmosphere_mod', &
                               'size of argument is incorrect', FATAL   )

   axes (1:size(axes(:))) = Atm(mytile)%atmos_axes (1:size(axes(:)))

 end subroutine get_atmosphere_axes



 subroutine get_bottom_mass ( t_bot, tr_bot, p_bot, z_bot, p_surf, slp )
!--------------------------------------------------------------
! returns temp, sphum, pres, height at the lowest model level
! and surface pressure
!--------------------------------------------------------------
   real, intent(out), dimension(isc:iec,jsc:jec):: t_bot, p_bot, z_bot, p_surf
   real, intent(out), optional, dimension(isc:iec,jsc:jec):: slp
   real, intent(out), dimension(isc:iec,jsc:jec,nq):: tr_bot
   integer :: i, j, m, k, kr
   real    :: rrg, sigtop, sigbot
   real, dimension(isc:iec,jsc:jec) :: tref
   real, parameter :: tlaps = 6.5e-3

   rrg  = rdgas / grav

   do j=jsc,jec
      do i=isc,iec
         p_surf(i,j) = Atm(mytile)%ps(i,j)
         t_bot(i,j) = Atm(mytile)%pt(i,j,npz)
         p_bot(i,j) = Atm(mytile)%delp(i,j,npz)/(Atm(mytile)%peln(i,npz+1,j)-Atm(mytile)%peln(i,npz,j))
         z_bot(i,j) = rrg*t_bot(i,j)*(1.+zvir*Atm(mytile)%q(i,j,npz,1)) *  &
                      (1. - Atm(mytile)%pe(i,npz,j)/p_bot(i,j))
      enddo
   enddo

   if ( present(slp) ) then
     ! determine 0.8 sigma reference level
     sigtop = Atm(mytile)%ak(1)/pstd_mks+Atm(mytile)%bk(1)
     do k = 1, npz
        sigbot = Atm(mytile)%ak(k+1)/pstd_mks+Atm(mytile)%bk(k+1)
        if (sigbot+sigtop > 1.6) then
           kr = k
           exit
        endif
        sigtop = sigbot
     enddo
     do j=jsc,jec
        do i=isc,iec
           ! sea level pressure
           tref(i,j) = Atm(mytile)%pt(i,j,kr) * (Atm(mytile)%delp(i,j,kr)/ &
                            ((Atm(mytile)%peln(i,kr+1,j)-Atm(mytile)%peln(i,kr,j))*Atm(mytile)%ps(i,j)))**(-rrg*tlaps)
           slp(i,j) = Atm(mytile)%ps(i,j)*(1.+tlaps*Atm(mytile)%phis(i,j)/(tref(i,j)*grav))**(1./(rrg*tlaps))
        enddo
     enddo
   endif

! Copy tracers
   do m=1,nq
      do j=jsc,jec
         do i=isc,iec
            tr_bot(i,j,m) = Atm(mytile)%q(i,j,npz,m)
         enddo
      enddo
   enddo

 end subroutine get_bottom_mass


 subroutine get_bottom_wind ( u_bot, v_bot )
!-----------------------------------------------------------
! returns u and v on the mass grid at the lowest model level
!-----------------------------------------------------------
   real, intent(out), dimension(isc:iec,jsc:jec):: u_bot, v_bot
   integer i, j

   do j=jsc,jec
      do i=isc,iec
         u_bot(i,j) = Atm(mytile)%u_srf(i,j)
         v_bot(i,j) = Atm(mytile)%v_srf(i,j)
      enddo
   enddo

 end subroutine get_bottom_wind



 subroutine get_stock_pe(index, value)
   integer, intent(in) :: index
   real,   intent(out) :: value

#ifdef USE_STOCK
   include 'stock.inc'
#endif

   real wm(isc:iec,jsc:jec)
   integer i,j,k
   real, pointer :: area(:,:)

   area => Atm(mytile)%gridstruct%area

   select case (index)

#ifdef USE_STOCK
   case (ISTOCK_WATER)
#else
   case (1)
#endif

!----------------------
! Perform vertical sum:
!----------------------
     wm = 0.
     do j=jsc,jec
        do k=1,npz
           do i=isc,iec
! Warning: the following works only with AM2 physics: water vapor; cloud water, cloud ice.
              wm(i,j) = wm(i,j) + Atm(mytile)%delp(i,j,k) * ( Atm(mytile)%q(i,j,k,1) +    &
                                                         Atm(mytile)%q(i,j,k,2) +    &
                                                         Atm(mytile)%q(i,j,k,3) )
           enddo
        enddo
     enddo

!----------------------
! Horizontal sum:
!----------------------
     value = 0.
     do j=jsc,jec
        do i=isc,iec
           value = value + wm(i,j)*area(i,j)
        enddo
     enddo
     value = value/grav

   case default
     value = 0.0
   end select

 end subroutine get_stock_pe


 subroutine atmosphere_state_update (Time, Physics_tendency, Physics, Atm_block)
   type(time_type),intent(in)      :: Time
   type (physics_tendency_type),   intent(in) :: Physics_tendency
   type (physics_type),    intent(in) :: Physics
   type (block_control_type), intent(in) :: Atm_block
   type(time_type) :: Time_prev, Time_next
!--- local variables ---
   integer :: i, j, k, n, w_diff, nt_dyn
   integer :: nb, ibs, ibe, jbs, jbe
   real ::  rcp

   Time_prev = Time
   Time_next = Time + Time_step_atmos

   n = mytile

   call set_domain ( Atm(mytile)%domain )

!--- put u/v tendencies into haloed arrays u_dt and v_dt
!$OMP parallel do default(shared) private(nb, ibs, ibe, jbs, jbe)
   do nb = 1,Atm_block%nblks
     ibs = Atm_block%ibs(nb)
     ibe = Atm_block%ibe(nb)
     jbs = Atm_block%jbs(nb)
     jbe = Atm_block%jbe(nb)

     u_dt(ibs:ibe,jbs:jbe,:)   = Physics_tendency%block(nb)%u_dt
     v_dt(ibs:ibe,jbs:jbe,:)   = Physics_tendency%block(nb)%v_dt
     t_dt(ibs:ibe,jbs:jbe,:)   = Physics_tendency%block(nb)%t_dt
     q_dt(ibs:ibe,jbs:jbe,:,:) = Physics_tendency%block(nb)%q_dt

!--- diagnostic tracers are being updated in-place
!--- tracer fields must be returned to the Atm structure
     Atm(mytile)%qdiag(ibs:ibe,jbs:jbe,:,:) = Physics_tendency%block(nb)%qdiag

   enddo

   w_diff = get_tracer_index (MODEL_ATMOS, 'w_diff' )
   nt_dyn = ncnst-pnats   !nothing more than nq
   if ( w_diff /= NO_TRACER ) then
      nt_dyn = nt_dyn - 1
   endif

!--- adjust w and heat tendency for non-hydrostatic case
    if ( .not.Atm(n)%flagstruct%hydrostatic .and. w_diff /= NO_TRACER ) then
      rcp = 1. / cp_air
!$OMP parallel do default (none) &
!$OMP              shared (jsc, jec, isc, iec, n, w_diff, Atm, q_dt, t_dt, rcp, dt_atmos) &
!$OMP             private (i, j, k)
       do k=1, Atm(n)%npz
         do j=jsc, jec
           do i=isc, iec
             Atm(n)%q(i,j,k,w_diff) = q_dt(i,j,k,w_diff) ! w tendency due to phys
! Heating due to loss of KE (vertical diffusion of w)
             t_dt(i,j,k) = t_dt(i,j,k) - q_dt(i,j,k,w_diff)*rcp*&
                                     (Atm(n)%w(i,j,k)+0.5*dt_atmos*q_dt(i,j,k,w_diff))
             Atm(n)%w(i,j,k) = Atm(n)%w(i,j,k) + dt_atmos*Atm(n)%q(i,j,k,w_diff)
           enddo
         enddo
       enddo
    endif

   call mpp_clock_begin (id_dynam)
       call timing_on('FV_UPDATE_PHYS')
    call fv_update_phys( dt_atmos, isc, iec, jsc, jec, isd, ied, jsd, jed, Atm(n)%ng, nt_dyn, &
                         Atm(n)%u,  Atm(n)%v,   Atm(n)%w,  Atm(n)%delp, Atm(n)%pt,         &
                         Atm(n)%q,  Atm(n)%qdiag,                                          &
                         Atm(n)%ua, Atm(n)%va,  Atm(n)%ps, Atm(n)%pe,   Atm(n)%peln,       &
                         Atm(n)%pk, Atm(n)%pkz, Atm(n)%ak, Atm(n)%bk,   Atm(n)%phis,       &
                         Atm(n)%u_srf, Atm(n)%v_srf, Atm(n)%ts, Atm(n)%delz,               &
                         Atm(n)%flagstruct%hydrostatic, u_dt, v_dt, t_dt,                  &
                         .true., Time_next, Atm(n)%flagstruct%nudge, Atm(n)%gridstruct,    &
                         Atm(n)%gridstruct%agrid(:,:,1), Atm(n)%gridstruct%agrid(:,:,2),   &
                         Atm(n)%npx, Atm(n)%npy, Atm(n)%npz, Atm(n)%flagstruct,            &
                         Atm(n)%neststruct, Atm(n)%bd, Atm(n)%domain, Atm(n)%ptop, Atm(n)%phys_diag, q_dt)
       call timing_off('FV_UPDATE_PHYS')
   call mpp_clock_end (id_dynam)

!--- nesting update after updating atmospheric variables with
!--- physics tendencies
    if (ngrids > 1 .and. p_split > 0) then
       call timing_on('TWOWAY_UPDATE')
       call twoway_nesting(Atm, ngrids, grids_on_this_pe, zvir, fv_time, mytile)
       call timing_off('TWOWAY_UPDATE')
    endif

#if !defined(ATMOS_NUDGE) && !defined(CLIMATE_NUDGE) && !defined(ADA_NUDGE)
   if ( .not.forecast_mode .and. Atm(mytile)%flagstruct%nudge .and. Atm(mytile)%flagstruct%na_init>0 ) then
        if(mod(seconds, 21600)==0)  call adiabatic_init_drv (Time_prev, Time_next)
   endif
#endif

   call nullify_domain()
  !---- diagnostics for FV dynamics -----
   if (Atm(mytile)%flagstruct%print_freq /= -99) then
     call mpp_clock_begin(id_fv_diag)
     call timing_on('FV_DIAG')

     fv_time = Time_next
     call get_time (fv_time, seconds,  days)

     call fv_diag(Atm(mytile:mytile), zvir, fv_time, Atm(mytile)%flagstruct%print_freq)
     call fv_cmip_diag(Atm(mytile:mytile), zvir, fv_time)

     call timing_off('FV_DIAG')
     call mpp_clock_end(id_fv_diag)
   endif

 end subroutine atmosphere_state_update

 subroutine adiabatic_init_drv (Time_prev, Time_next)
   type (time_type),       intent(in) :: Time_prev, Time_next
   integer:: n
   real, allocatable, dimension(:,:,:):: u_dt, v_dt, t_dt
   real, allocatable:: q_dt(:,:,:,:)
   integer:: isd, ied, jsd, jed, ngc

   character(len=80) :: errstr

!---------------------------------------------------
! Call the adiabatic forward-backward initialization
!---------------------------------------------------
   write(errstr,'(A, I4, A)') 'Performing adiabatic nudging',  Atm(mytile)%flagstruct%na_init, ' times'
   call mpp_error(NOTE, errstr)

        ngc = Atm(mytile)%ng
        isd = isc - ngc
        ied = iec + ngc
        jsd = jsc - ngc
        jed = jec + ngc

     call timing_on('adiabatic_init')

        allocate ( u_dt(isd:ied,jsd:jed, npz) )
        allocate ( v_dt(isd:ied,jsd:jed, npz) )
        allocate ( t_dt(isc:iec,jsc:jec, npz) )
        allocate ( q_dt(isc:iec,jsc:jec, npz, nq) )

     do_adiabatic_init = .true.

     do n=1,Atm(mytile)%flagstruct%na_init
        call adiabatic_init(Atm, Time_next, -dt_atmos, u_dt, v_dt, t_dt, q_dt, .false.)  ! Backward in time one step
        fv_time = Time_prev
        call adiabatic_init(Atm, Time_next,  dt_atmos, u_dt, v_dt, t_dt, q_dt, .true. )  ! Forward to the original time
        fv_time = Time_next
     enddo

     do_adiabatic_init = .false.

        deallocate ( u_dt )
        deallocate ( v_dt )
        deallocate ( t_dt )
        deallocate ( q_dt )

     call timing_off('adiabatic_init')

 end subroutine adiabatic_init_drv

 subroutine adiabatic_init (Atm, Time, dt_init, u_dt, v_dt, t_dt, q_dt, do_nudge)
!
! Perform adiabatic forward-backward adiabatic initialization with nudging
!
    type(fv_atmos_type), intent(inout) :: Atm(:)
    type (time_type),    intent(in) :: Time
    real, intent(in):: dt_init
    logical, intent(in):: do_nudge
    real, intent(inout), dimension(:,:,:):: u_dt, v_dt, t_dt
    real, intent(inout)::  q_dt(:,:,:,:)
!
    integer:: isd, ied, jsd, jed, ngc, n
    type(time_type) :: Time_next

    Time_next = Time + Time_step_atmos

    n = mytile
    ngc = Atm(mytile)%ng
    isd = isc - ngc
    ied = iec + ngc
    jsd = jsc - ngc
    jed = jec + ngc

    call fv_dynamics(npx, npy, npz, nq, ngc,  dt_init, 0.,                            &
                     .false.,  Atm(n)%flagstruct%reproduce_sum, kappa, cp_air, zvir,  &
                     Atm(n)%ptop, Atm(n)%ks, ncnst, Atm(n)%flagstruct%n_split,        &
                     Atm(n)%flagstruct%q_split, Atm(n)%u, Atm(n)%v, Atm(n)%w,         &
                     Atm(n)%delz, Atm(n)%flagstruct%hydrostatic,                      &
                     Atm(n)%pt, Atm(n)%delp, Atm(n)%q, Atm(n)%ps,                     &
                     Atm(n)%pe, Atm(n)%pk, Atm(n)%peln, Atm(n)%pkz, Atm(n)%phis,      &
                     Atm(n)%q_con, Atm(n)%omga, Atm(n)%ua, Atm(n)%va, Atm(n)%uc, Atm(n)%vc,  &
                     Atm(n)%ak, Atm(n)%bk, Atm(n)%mfx, Atm(n)%mfy,                    &
                     Atm(n)%cx, Atm(n)%cy, Atm(n)%ze0, Atm(n)%flagstruct%hybrid_z,    &
                     Atm(n)%gridstruct, Atm(n)%flagstruct,                            &
                     Atm(n)%neststruct, Atm(n)%idiag, Atm(n)%bd,                      &
                     Atm(n)%parent_grid, Atm(n)%domain, Atm(n)%inline_mp)

! No large-scale nudging at "Time_prev"
    if ( do_nudge ) then
         u_dt = 0.  ;   v_dt = 0. ;  t_dt = 0. ;  q_dt = 0.
    call fv_update_phys( abs(dt_init), isc, iec, jsc, jec, isd, ied, jsd, jed, ngc, nq,  &
                         Atm(n)%u,  Atm(n)%v,   Atm(n)%w,  Atm(n)%delp, Atm(n)%pt,       &
                         Atm(n)%q,  Atm(n)%qdiag,                                        &
                         Atm(n)%ua, Atm(n)%va,  Atm(n)%ps, Atm(n)%pe,   Atm(n)%peln,     &
                         Atm(n)%pk, Atm(n)%pkz, Atm(n)%ak, Atm(n)%bk,   Atm(n)%phis,     &
                         Atm(n)%u_srf, Atm(n)%v_srf, Atm(n)%ts, Atm(n)%delz,             &
                         Atm(n)%flagstruct%hydrostatic, u_dt, v_dt, t_dt,                &
                         .true., Time_next, Atm(n)%flagstruct%nudge, Atm(n)%gridstruct,  &
                         Atm(n)%gridstruct%agrid(:,:,1), Atm(n)%gridstruct%agrid(:,:,2), &
                         Atm(n)%npx, Atm(n)%npy, Atm(n)%npz, Atm(n)%flagstruct,          &
                         Atm(n)%neststruct, Atm(n)%bd, Atm(n)%domain, Atm(n)%ptop, q_dt)

    endif

 end subroutine adiabatic_init

 subroutine atmos_physics_driver_inputs (Physics, Atm_block, Physics_tendency)
   type (physics_type),  intent(inout) :: Physics
   type (block_control_type), intent(in) :: Atm_block
   type (physics_tendency_type), intent(inout), optional :: Physics_tendency
!--- local variabls
   integer :: nb, ibs, ibe, jbs, jbe

!---------------------------------------------------------------------
! use most up to date atmospheric properties when running serially
!---------------------------------------------------------------------
!$OMP parallel do default(shared) private(nb, ibs, ibe, jbs, jbe)
   do nb = 1, Atm_block%nblks
     ibs = Atm_block%ibs(nb)
     ibe = Atm_block%ibe(nb)
     jbs = Atm_block%jbs(nb)
     jbe = Atm_block%jbe(nb)

     Physics%block(nb)%phis = Atm(mytile)%phis(ibs:ibe,jbs:jbe)
     Physics%block(nb)%u    = Atm(mytile)%ua(ibs:ibe,jbs:jbe,:)
     Physics%block(nb)%v    = Atm(mytile)%va(ibs:ibe,jbs:jbe,:)
     Physics%block(nb)%t    = Atm(mytile)%pt(ibs:ibe,jbs:jbe,:)
     Physics%block(nb)%q    = Atm(mytile)%q(ibs:ibe,jbs:jbe,:,:)
     Physics%block(nb)%omega= Atm(mytile)%omga(ibs:ibe,jbs:jbe,:)
     Physics%block(nb)%pe   = Atm(mytile)%pe(ibs:ibe,:,jbs:jbe)
     Physics%block(nb)%peln = Atm(mytile)%peln(ibs:ibe,:,jbs:jbe)
     Physics%block(nb)%delp = Atm(mytile)%delp(ibs:ibe,jbs:jbe,:)
     if (.not.Physics%control%phys_hydrostatic) then
        Physics%block(nb)%delz = Atm(mytile)%delz(ibs:ibe,jbs:jbe,:)
        Physics%block(nb)%w    = Atm(mytile)%w(ibs:ibe,jbs:jbe,:)
     endif
     if (_ALLOCATED(Physics%block(nb)%tmp_4d)) &
        Physics%block(nb)%tmp_4d = Atm(mytile)%qdiag(ibs:ibe,jbs:jbe,:,:)

     call fv_compute_p_z (Atm_block%npz, Physics%block(nb)%phis, Physics%block(nb)%pe, &
                          Physics%block(nb)%peln, Physics%block(nb)%delp, Physics%block(nb)%delz, &
                          Physics%block(nb)%t, Physics%block(nb)%q(:,:,:,Physics%control%sphum), &
                          Physics%block(nb)%p_full, Physics%block(nb)%p_half, &
                          Physics%block(nb)%z_full, Physics%block(nb)%z_half, &
#ifdef USE_COND
                          Atm(mytile)%q_con(ibs:ibe,jbs:jbe,:), &
#else
                          Atm(mytile)%q_con, &
#endif
                          Physics%control%phys_hydrostatic, Physics%control%do_uni_zfull) !miz

     if (PRESENT(Physics_tendency)) then
!--- copy the dynamics tendencies into the physics tendencies
!--- if one wants to run physics concurrent with dynamics,
!--- these values would be zeroed out and accumulated
!--- in the atmosphere_state_update

       Physics_tendency%block(nb)%u_dt = u_dt(ibs:ibe,jbs:jbe,:)
       Physics_tendency%block(nb)%v_dt = v_dt(ibs:ibe,jbs:jbe,:)
       Physics_tendency%block(nb)%t_dt = t_dt(ibs:ibe,jbs:jbe,:)
       Physics_tendency%block(nb)%q_dt = q_dt(ibs:ibe,jbs:jbe,:,:)
       Physics_tendency%block(nb)%qdiag = Atm(mytile)%qdiag(ibs:ibe,jbs:jbe,:,:)
     endif
   enddo

 end subroutine atmos_physics_driver_inputs


 subroutine atmos_radiation_driver_inputs (Time, Radiation, Atm_block)
   type (time_type),          intent(in)    :: Time
   type (radiation_type),     intent(inout) :: Radiation
   type (block_control_type), intent(in)    :: Atm_block
!--- local variables
   integer :: nb, ibs, ibe, jbs, jbe

!---------------------------------------------------------------------
! use most up to date atmospheric properties when running serially
!---------------------------------------------------------------------
!$OMP parallel do default(shared) private(nb, ibs, ibe, jbs, jbe)
   do nb = 1,Atm_block%nblks
     ibs = Atm_block%ibs(nb)
     ibe = Atm_block%ibe(nb)
     jbs = Atm_block%jbs(nb)
     jbe = Atm_block%jbe(nb)

     Radiation%block(nb)%phis = Atm(mytile)%phis(ibs:ibe,jbs:jbe)
     Radiation%block(nb)%t    = Atm(mytile)%pt(ibs:ibe,jbs:jbe,:)
     Radiation%block(nb)%q    = Atm(mytile)%q(ibs:ibe,jbs:jbe,:,:)
     Radiation%block(nb)%pe   = Atm(mytile)%pe(ibs:ibe,:,jbs:jbe)
     Radiation%block(nb)%peln = Atm(mytile)%peln(ibs:ibe,:,jbs:jbe)
     Radiation%block(nb)%delp = Atm(mytile)%delp(ibs:ibe,jbs:jbe,:)
     if (.not.Radiation%control%phys_hydrostatic) &
        Radiation%block(nb)%delz = Atm(mytile)%delz(ibs:ibe,jbs:jbe,:)

     call fv_compute_p_z (Atm_block%npz, Radiation%block(nb)%phis, Radiation%block(nb)%pe, &
                          Radiation%block(nb)%peln, Radiation%block(nb)%delp, Radiation%block(nb)%delz, &
                          Radiation%block(nb)%t, Radiation%block(nb)%q(:,:,:,Radiation%control%sphum), &
                          Radiation%block(nb)%p_full, Radiation%block(nb)%p_half, &
                          Radiation%block(nb)%z_full, Radiation%block(nb)%z_half, &
#ifdef USE_COND
                          Atm(mytile)%q_con(ibs:ibe,jbs:jbe,:), &
#else
                          Atm(mytile)%q_con, &
#endif
                          Radiation%control%phys_hydrostatic, Radiation%control%do_uni_zfull) !miz
   enddo

!----------------------------------------------------------------------
! obtain pressure-weighted global mean co2 dry volume mixing ratio for
! use by radiation package.
!----------------------------------------------------------------------
! compute_g_avg must be called here because it contains
! mpp_sums that cannot be called during the concurrent radiation
! phase due to the way in which MPI interacts with nested OpenMP
!----------------------------------------------------------------------
   call compute_g_avg(Time, 'co2', Radiation, Atm_block)

 end subroutine atmos_radiation_driver_inputs



 subroutine fv_compute_p_z (npz, phis, pe, peln, delp, delz, pt, q_sph, &
                            p_full, p_half, z_full, z_half, q_con, hydrostatic, do_uni_zfull) !miz
    integer, intent(in)  :: npz
    real, dimension(:,:),   intent(in)  :: phis
    real, dimension(:,:,:), intent(in)  :: pe, peln, delp, delz, q_con, pt, q_sph
    real, dimension(:,:,:), intent(out) :: p_full, p_half, z_full, z_half
    logical, intent(in)  :: hydrostatic, do_uni_zfull !miz
!--- local variables
    integer i,j,k,isiz,jsiz
    real    tvm
    real    :: zvir, rrg, ginv
#ifdef USE_COND
    real, dimension(size(pe,1),size(pe,3),size(pe,2):: peg, pelng
#endif

    isiz=size(phis,1)
    jsiz=size(phis,2)
    zvir = rvgas/rdgas - 1.
    ginv = 1./ grav
    rrg  = rdgas / grav

!----------------------------------------------------
! Compute pressure and height at full and half levels
!----------------------------------------------------
    z_half(:,:,npz+1) = phis(:,:) * ginv

    do k=1,npz+1
      do j=1,jsiz
         do i=1,isiz
           p_half(i,j,k) = pe(i,k,j)
        enddo
      enddo
    enddo

!--------- Hydrostatic option ----------------------------------------------
    if (hydrostatic ) then
#ifdef USE_COND
    do j=1,jsiz
       do i=1,isiz
          peg(i,j,1) = pe(i,1,j)
       enddo
    end do
    do k=2,npz+1
       do j=1,jsiz
          do i=1,isiz
             peg(i,j,k) = peg(i,j,k-1) + delp(i,j,k-1)*(1.-q_con(i,j,k-1))
          enddo
       enddo
    enddo
#endif
      do k=npz,1,-1
        do j=1,jsiz
          do i=1,isiz
            tvm = rrg*pt(i,j,k)*(1.+zvir*q_sph(i,j,k))
            p_full(i,j,k) = delp(i,j,k)/(peln(i,k+1,j)-peln(i,k,j))
#ifdef USE_COND
            dlg = log(peg(i,j,k+1)/peg(i,j,k))
            z_full(i,j,k) = z_half(i,j,k+1) + tvm*(1.-peg(i,j,k)*dlg/(peg(i,j,k+1)-peg(i,j,k)))
            z_half(i,j,k) = z_half(i,j,k+1) + tvm*dlg
#else
            z_full(i,j,k) = z_half(i,j,k+1) + tvm*(1.-p_half(i,j,k)/p_full(i,j,k))
            z_half(i,j,k) = z_half(i,j,k+1) + tvm*(peln(i,k+1,j)-peln(i,k,j))
#endif
          enddo
        enddo
      enddo
    else
!--------- Non-Hydrostatic option ------------------------------------------
      do k=npz,1,-1
        do j=1,jsiz
          do i=1,isiz
            p_full(i,j,k) = delp(i,j,k)/(peln(i,k+1,j)-peln(i,k,j))
            z_half(i,j,k) = z_half(i,j,k+1) - delz(i,j,k)
            z_full(i,j,k) = 0.5*(z_half(i,j,k) + z_half(i,j,k+1))
          enddo
        enddo
      enddo
    endif
    if (do_uni_zfull) then
       do k=1,npz
          z_full(:,:,k)=0.5*(z_half(:,:,k)+z_half(:,:,k+1))
       enddo
    endif
  end subroutine fv_compute_p_z


 subroutine reset_atmos_tracers (Physics, Physics_tendency, Atm_block)
   type (physics_type), intent(in) :: Physics
   type (physics_tendency_type), intent(in) :: Physics_tendency
   type (block_control_type), intent(in) :: Atm_block
!--- local variables
   integer :: nb, ibs, ibe, jbs, jbe

!--- After initialization by the physics, tracer fields must be
!--- returned to the Atm structure.  This is because tracer driver
!--- can reset the initial values
!$OMP parallel do default(shared) private(nb, ibs, ibe, jbs, jbe)
    do nb = 1, Atm_block%nblks
      ibs = Atm_block%ibs(nb)
      ibe = Atm_block%ibe(nb)
      jbs = Atm_block%jbs(nb)
      jbe = Atm_block%jbe(nb)

      Atm(mytile)%q(ibs:ibe,jbs:jbe,:,:) = Physics%block(nb)%q
      Atm(mytile)%qdiag(ibs:ibe,jbs:jbe,:,:) = Physics_tendency%block(nb)%qdiag
    enddo

 end subroutine reset_atmos_tracers

end module atmosphere_mod
